# Copyright 2019 Kemal Kurniawan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque
from typing import Any, Callable, Deque, Optional
from pathlib import Path
import logging
import pickle

from .runner import Runner

logger = logging.getLogger(__name__)


class EarlyStopper:
    """A handler for early stopping.

    This handler keeps track the number of times the loss value does not improve. If this
    number is greater than the given patience, this handler stops the given runner.

    Example:

        >>> valid_losses = [0.1, 0.2, 0.3]  # simulate validation batch losses
        >>> dummy_batches = range(10)
        >>> dummy_batch_fn = lambda x: x
        >>>
        >>> from rnnr import Event, Runner
        >>> from rnnr.attachments import MeanAggregator
        >>> from rnnr.handlers import EarlyStopper
        >>>
        >>> trainer, evaluator = Runner(), Runner()
        >>> @trainer.on(Event.EPOCH_STARTED)
        ... def print_epoch(state):
        ...     print('Epoch', state['epoch'], 'started')
        ...
        >>> @trainer.on(Event.EPOCH_FINISHED)
        ... def eval_on_valid(state):
        ...     evaluator.run(lambda loss: loss, valid_losses)
        ...
        >>> MeanAggregator(name='loss').attach_on(evaluator)
        >>> evaluator.append_handler(Event.FINISHED, EarlyStopper(trainer, patience=2))
        >>> _ = trainer.run(dummy_batch_fn, dummy_batches, max_epoch=7)
        Epoch 1 started
        Epoch 2 started
        Epoch 3 started
        Epoch 4 started

    Args:
        runner: Runner to stop early.
        patience: Number of times to wait for the loss to improve before stopping.
        loss_fn: Callback to get the loss value from the runner's ``state`` on which this
            handler is appended. The default is to get ``state['loss']`` as the loss.
        eps: An improvement is considered only when the loss value decreases by at least
            this amount.
    """

    def __init__(
            self,
            runner: Runner,
            patience: int = 5,
            loss_fn: Optional[Callable[[dict], float]] = None,
            eps: float = 1e-4,
    ) -> None:
        if loss_fn is None:
            loss_fn = lambda state: state['loss']

        self._runner = runner
        self._patience = patience
        self._loss_fn = loss_fn
        self._eps = eps

        self._num_bad_loss = 0
        self._min_loss = float('inf')

    def __call__(self, state: dict) -> None:
        loss = self._loss_fn(state)
        if loss <= self._min_loss - self._eps:
            self._min_loss = loss
            self._num_bad_loss = 0
        else:
            self._num_bad_loss += 1

        if self._num_bad_loss > self._patience:
            logger.info('Patience exceeded, stopping early')
            self._runner.stop()


class Checkpointer:
    """A handler for checkpointing.

    Checkpointing here means saving some objects (e.g., models) periodically during a run.

    Example:

        >>> from pathlib import Path
        >>> from pprint import pprint
        >>> from rnnr import Event, Runner
        >>> from rnnr.handlers import Checkpointer
        >>>
        >>> dummy_batches = range(3)
        >>> dummy_batch_fn = lambda x: x
        >>> tmp_dir = Path('/tmp')
        >>> objs = {'model.pkl': 'MODEL', 'optimizer.pkl': 'OPTIMIZER'}
        >>> runner = Runner()
        >>> runner.append_handler(Event.EPOCH_FINISHED, Checkpointer(tmp_dir, objs, max_saved=3))
        >>> _ = runner.run(dummy_batch_fn, dummy_batches, max_epoch=7)
        >>> pprint(sorted(list(tmp_dir.glob('*.pkl'))))
        [PosixPath('/tmp/5_model.pkl'),
         PosixPath('/tmp/5_optimizer.pkl'),
         PosixPath('/tmp/6_model.pkl'),
         PosixPath('/tmp/6_optimizer.pkl'),
         PosixPath('/tmp/7_model.pkl'),
         PosixPath('/tmp/7_optimizer.pkl')]

    Args:
        save_dir: Save checkpoints in this directory.
        objs: Dictionary whose keys are filenames and the values are the objects to checkpoint.
            The filenames in this dictionary's keys are prepended with the number of times
            this handler is called to get the actual saved files' names. This allows the
            actual filenames contain the e.g. epoch number if this handler is invoked at the
            end of each epoch.
        max_saved: Maximum number of checkpoints saved.
        loss_fn: If given, this should return the loss value of the given runner's state dict.
            Checkpoints are saved only when the returned loss is smaller than the minimum loss
            observed so far. The default of ``None`` means checkpoints are saved whenever this
            handler is called.
        save_fn: Function to invoke to save the checkpoints. If given, this must be a callable
            accepting two arguments: an object to save and a path to save it to. The default
            is to save the object using `pickle`.
        eps: The loss value must be smaller at least by this value to be considered as an
            improvement. Only used if ``loss_fn`` is given.
    """

    def __init__(
            self,
            save_dir: Path,
            objs: dict,
            max_saved: int = 1,
            loss_fn: Optional[Callable[[dict], float]] = None,
            save_fn: Optional[Callable[[Any, Path], None]] = None,
            eps: float = 1e-4,
    ) -> None:
        if save_fn is None:
            save_fn = self._default_save_fn

        self._save_dir = save_dir
        self._objs = objs
        self._max_saved = max_saved
        self._loss_fn = loss_fn
        self._save_fn = save_fn
        self._eps = eps

        self._num_calls = 0
        self._deque: Deque[int] = deque()
        self._min_loss = float('inf')

    @staticmethod
    def _default_save_fn(obj: Any, path: Path) -> None:
        with open(path, 'wb') as f:
            pickle.dump(obj, f)

    def __call__(self, state: dict) -> None:
        self._num_calls += 1
        if self._should_save(state):
            self._save()
        if self._should_delete():
            self._delete()

        assert self._num_saved <= self._max_saved

    @property
    def _num_saved(self) -> int:
        return len(self._deque)

    def _should_save(self, state: dict) -> bool:
        if self._loss_fn is None:
            return True

        loss = self._loss_fn(state)
        if loss <= self._min_loss - self._eps:
            logger.info('Found new best loss of %f', loss)
            self._min_loss = loss
            return True

        return False

    def _should_delete(self) -> bool:
        return self._num_saved > self._max_saved

    def _save(self) -> None:
        self._deque.append(self._num_calls)
        for name, obj in self._objs.items():
            path = self._save_dir / f'{self._num_calls}_{name}'
            logger.info('Saving to %s', path)
            self._save_fn(obj, path)

    def _delete(self) -> None:
        num = self._deque.popleft()
        for name in self._objs:
            path = self._save_dir / f'{num}_{name}'
            if path.exists():
                path.unlink()



from unittest.mock import Mock, call
import pickle

from rnnr.handlers import Checkpointer


def test_ok(tmp_path):
    num_calls = 5
    objs_values = {
        'model.pkl': [f'MODEL_{i}' for i in range(num_calls)],
        'opt.pkl': [f'OPT_{i}' for i in range(num_calls)],
    }
    objs = {'model.pkl': None, 'opt.pkl': None}

    ckptr = Checkpointer(tmp_path, objs)
    for i in range(num_calls):
        for name in objs:
            objs[name] = objs_values[name][i]
        ckptr({})

    for name in objs:
        path = tmp_path / f'{num_calls}_{name}'
        with open(path, 'rb') as f:
            assert pickle.load(f) == objs_values[name][-1]


def test_max_saved(tmp_path):
    num_calls, max_saved = 5, 3
    objs_values = {
        'model.pkl': [f'MODEL_{i}' for i in range(num_calls)],
        'opt.pkl': [f'OPT_{i}' for i in range(num_calls)],
    }
    objs = {'model.pkl': None, 'opt.pkl': None}

    ckptr = Checkpointer(tmp_path, objs, max_saved=max_saved)
    for i in range(num_calls):
        for name in objs:
            objs[name] = objs_values[name][i]
        ckptr({})

    for name in objs:
        assert len(list(tmp_path.glob(f'*_{name}'))) == max_saved
        for i in range(max_saved):
            path = tmp_path / f'{num_calls - i}_{name}'
            with open(path, 'rb') as f:
                assert pickle.load(f) == objs_values[name][num_calls - i - 1]


def test_loss_fn(tmp_path):
    num_calls, max_saved = 5, 2
    objs_values = {
        'model.pkl': [f'MODEL_{i}' for i in range(num_calls)],
        'opt.pkl': [f'OPT_{i}' for i in range(num_calls)],
    }
    objs = {'model.pkl': None, 'opt.pkl': None}
    # new best losses are call #1, #3, and #5
    losses = [3, 4, 2, 4, 1]
    loss_fn = lambda state: losses[state['i']]

    ckptr = Checkpointer(tmp_path, objs, max_saved=max_saved, loss_fn=loss_fn)
    for i in range(num_calls):
        for name in objs:
            objs[name] = objs_values[name][i]
        ckptr({'i': i})

    # saved call are #3 and #5 (the last 2)
    saved_calls = [3, 5]
    for name in objs:
        assert len(list(tmp_path.glob(f'*_{name}'))) == max_saved
        for c in saved_calls:
            path = tmp_path / f'{c}_{name}'
            with open(path, 'rb') as f:
                assert pickle.load(f) == objs_values[name][c - 1]


def test_save_fn(tmp_path):
    num_calls = 5
    objs_values = {
        'model.pkl': [f'MODEL_{i}' for i in range(num_calls)],
        'opt.pkl': [f'OPT_{i}' for i in range(num_calls)],
    }
    objs = {'model.pkl': None, 'opt.pkl': None}
    mock_save_fn = Mock()

    ckptr = Checkpointer(tmp_path, objs, save_fn=mock_save_fn)
    for i in range(num_calls):
        for name in objs:
            objs[name] = objs_values[name][i]
        ckptr({})

    assert mock_save_fn.mock_calls == [
        call(objs_values[name][i], tmp_path / f'{i+1}_{name}')
        for i in range(num_calls)
        for name in objs
    ]



import statistics as stat

import pytest

from rnnr import Event
from rnnr.attachments import MeanAggregator


def test_ok(runner):
    batches = range(5)
    values = [12, 7, 8, 44, -13]
    batch_fn = lambda b: values[b]

    def efhandler(state):
        assert state[agg.name] == pytest.approx(stat.mean(batch_fn(b) for b in batches))

    agg = MeanAggregator()
    assert agg.name == 'mean'
    agg.attach_on(runner)
    runner.append_handler(Event.EPOCH_FINISHED, efhandler)
    runner.run(batch_fn, batches)


def test_more_than_one_epoch(runner):
    batches = range(5)
    values = [12, 7, 8, 44, -13, 78, 55, -109, 34, 10]
    epoch = 0
    batch_fn = lambda b: (epoch - 1) * len(batches) + values[b]

    def eshandler(state):
        nonlocal epoch
        epoch = state['epoch']

    def efhandler(state):
        assert state[agg.name] == pytest.approx(stat.mean(batch_fn(b) for b in batches))

    agg = MeanAggregator()
    agg.attach_on(runner)
    runner.append_handler(Event.EPOCH_STARTED, eshandler)
    runner.append_handler(Event.EPOCH_FINISHED, efhandler)
    runner.run(batch_fn, batches, max_epoch=2)


def test_custom_name(runner):
    batches = range(10)
    batch_fn = lambda b: b**2

    def efhandler(state):
        assert state[agg.name] == pytest.approx(stat.mean(batch_fn(b) for b in batches))

    agg = MeanAggregator(name='avg')
    agg.attach_on(runner)
    runner.append_handler(Event.EPOCH_FINISHED, efhandler)
    runner.run(batch_fn, batches)


def test_value_fn(runner):
    batches = range(10)
    batch_fn = lambda b: (b**2, b**3)
    value_fn = lambda state: state['output'][1]

    def efhandler(state):
        assert state[agg.name] == pytest.approx(stat.mean(b**3 for b in batches))

    agg = MeanAggregator(value_fn=value_fn)
    agg.attach_on(runner)
    runner.append_handler(Event.EPOCH_FINISHED, efhandler)
    runner.run(batch_fn, batches)


def test_size_fn(runner):
    batches = range(5)
    sizes = [3, 4, 9, 10, 2]
    batch_fn = lambda b: b
    size_fn = lambda state: sizes[state['batch']]

    def efhandler(state):
        assert state[agg.name] == pytest.approx(sum(batch_fn(b) for b in batches) / sum(sizes))

    agg = MeanAggregator(size_fn=size_fn)
    agg.attach_on(runner)
    runner.append_handler(Event.EPOCH_FINISHED, efhandler)
    runner.run(batch_fn, batches)



# Copyright 2019 Kemal Kurniawan
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Optional, Type
import abc

from tqdm import tqdm

from .event import Event
from .runner import Runner


class Attachment(abc.ABC):
    """An abstract base class for an attachment."""

    @abc.abstractmethod
    def attach_on(self, runner: Runner) -> None:
        """Attach to the given runner.

        Args:
            runner: Runner to attach to.
        """
        pass


class ProgressBar(Attachment):
    """An attachment to display a progress bar.

    The progress bar is implemented using `tqdm`_.

    Example:

        >>> from rnnr import Runner
        >>> from rnnr.attachments import ProgressBar
        >>> runner = Runner()
        >>> ProgressBar().attach_on(runner)
        >>> _ = runner.run(lambda x: x, range(10), max_epoch=10)

    Args:
        size_fn: Function to get the size of a batch to update the progress bar with.
            If not given, the default is to always return 1 as the size of a batch.
        stats_fn: Function to get the statistics dictionary to be displayed along with the
            progress bar. If given, it should accept a runner's state dictionary and return
            another dictionary whose keys are the names of the statistics and the values are
            the statistics values.
        **kwargs: Keyword arguments to be passed to `tqdm`_ class.


    .. _tqdm: https://github.com/tqdm/tqdm
    """

    def __init__(
            self,
            size_fn: Optional[Callable[[dict], int]] = None,
            stats_fn: Optional[Callable[[dict], dict]] = None,
            tqdm_cls: Optional[Type[tqdm]] = None,
            **kwargs,
    ) -> None:
        if tqdm_cls is None:  # pragma: no cover
            tqdm_cls = tqdm
        if size_fn is None:
            size_fn = lambda _: 1

        self._tqdm_cls = tqdm_cls
        self._size_fn = size_fn
        self._stats_fn = stats_fn
        self._kwargs = kwargs

        self._pbar: tqdm

    def attach_on(self, runner: Runner) -> None:
        runner.append_handler(Event.EPOCH_STARTED, self._create)
        runner.append_handler(Event.BATCH_FINISHED, self._update)
        runner.append_handler(Event.EPOCH_FINISHED, self._close)

    def _create(self, state: dict) -> None:
        self._pbar = self._tqdm_cls(state['batches'], **self._kwargs)

    def _update(self, state: dict) -> None:
        if self._stats_fn is not None:
            self._pbar.set_postfix(**self._stats_fn(state))
        self._pbar.update(self._size_fn(state))

    def _close(self, state: dict) -> None:
        self._pbar.close()


class MeanAggregator(Attachment):
    """An attachment to compute a mean over batch statistics.

    This attachment gets the value from each batch and compute their mean at the end of
    every epoch.

    Example:

        >>> from rnnr import Event, Runner
        >>> from rnnr.attachments import MeanAggregator
        >>> runner = Runner()
        >>> MeanAggregator().attach_on(runner)
        >>> runner.run(lambda x: x, [1, 2, 3])
        {'max_epoch': 1, 'batches': [1, 2, 3], 'mean': 2.0}

    Args:
        name: Name of this aggregator. This name is used as the key in the runner's state
            dictionary.
        value_fn: Function to get the value of a batch. If given, it should accept the
            runner's state dictionary at the end of a batch and return a value. The default
            is to get ``state['output']`` as the value.
        size_fn: Function to get the size of a batch. If given, it should accept the runner's
            state dictionary at the end of a batch and return the batch size. The default is
            to always return 1 as the batch size. The sum of all these batch sizes is the
            divisor when computing the mean.
    """

    def __init__(
            self,
            name: str = 'mean',
            value_fn: Optional[Callable[[dict], Any]] = None,
            size_fn: Optional[Callable[[dict], int]] = None,
    ) -> None:
        if value_fn is None:
            value_fn = lambda state: state['output']
        if size_fn is None:
            size_fn = lambda _: 1

        self.name = name
        self._value_fn = value_fn
        self._size_fn = size_fn

        self._total = 0
        self._size = 0

    def attach_on(self, runner: Runner) -> None:
        runner.append_handler(Event.EPOCH_STARTED, self._reset)
        runner.append_handler(Event.BATCH_FINISHED, self._update)
        runner.append_handler(Event.EPOCH_FINISHED, self._compute)

    def _reset(self, state: dict) -> None:
        self._total = 0
        self._size = 0

    def _update(self, state: dict) -> None:
        self._total += self._value_fn(state)
        self._size += self._size_fn(state)

    def _compute(self, state: dict) -> None:
        state[self.name] = self._total / self._size
